/*
    This file is part of CrabEmu.

    Copyright (C) 2007, 2008 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdlib.h>
#include <stdio.h>

#include <dc/pvr.h>
#include <dc/maple.h>
#include <dc/maple/controller.h>
#include <kos/sem.h>
#include <dc/video.h>
#include <dc/flashrom.h>

#include "CrabEmu.h"
#include "sms.h"
#include "smsvdp.h"
#include "smsz80.h"
#include "sn76489.h"
#include "smsmem.h"
#include "sound.h"

pvr_ptr_t tex[2];
pvr_ptr_t txr;
pvr_poly_hdr_t hdr[2];
int active_tex;
int region;
static semaphore_t *dma_sem;

extern uint16 sms_pad;
extern int sms_console;
extern int sms_region;
extern sn76489_t psg;
extern int sms_console;
extern int sms_psg_enabled;
extern uint32 psg_samples[313];
extern int sms_cycles_run;
extern int sms_cycles_to_run;

extern const char *menu(const char *dir);
extern int pal_menu(void);

static const float NTSC_Z80_CLOCK = 3579545.0f;
static const int PSG_DIVISOR = 16;
static const int NTSC_FPS = 60;
static const int NTSC_LINES_PER_FRAME = 262;
static const float NTSC_CLOCKS_PER_SAMPLE = 5.073051303855f;
static const int NTSC_SAMPLES_PER_FRAME = 735;

static const float PAL_Z80_CLOCK = 3546893.0f;
static const int PAL_FPS = 50;
static const int PAL_LINES_PER_FRAME = 313;
static const float PAL_CLOCKS_PER_SAMPLE = 5.02677579365f;
static const int PAL_SAMPLES_PER_FRAME = 882;

void dma_done(ptr_t data __UNUSED__)    {
    sem_signal(dma_sem);
}

void render_scene(void) {
    pvr_vertex_t vert;

    sem_wait(dma_sem);

    pvr_wait_ready();
    pvr_scene_begin();

    pvr_list_begin(PVR_LIST_OP_POLY);

    pvr_prim(hdr + active_tex, sizeof(pvr_poly_hdr_t));

    vert.flags = PVR_CMD_VERTEX;
    vert.argb = 0xFFFFFFFF;
    vert.oargb = 0;

    if(sms_console != CONSOLE_GG)   {
        float y = (480.0f - (smsvdp.lines << 1)) / 2.0f;

        vert.x = 64.0f;
        vert.y = y + (smsvdp.lines << 1);
        vert.z = 1.0f;
        vert.u = 0.0f;
        vert.v = smsvdp.lines / 256.0f;
        pvr_prim(&vert, sizeof(pvr_vertex_t));

        vert.y = y;
        vert.v = 0.0f;
        pvr_prim(&vert, sizeof(pvr_vertex_t));

        vert.x = 576.0f;
        vert.y = y + (smsvdp.lines << 1);
        vert.u = 1.0f;
        vert.v = smsvdp.lines / 256.0f;
        pvr_prim(&vert, sizeof(pvr_vertex_t));

        vert.flags = PVR_CMD_VERTEX_EOL;
        vert.y = y;
        vert.v = 0.0f;
        pvr_prim(&vert, sizeof(pvr_vertex_t));
    }
    else    {
        vert.x = 80.0f;
        vert.y = 456.0f;
        vert.z = 1.0f;
        vert.u = 48.0f / 256.0f;
        vert.v = 168.0f / 256.0f;
        pvr_prim(&vert, sizeof(pvr_vertex_t));

        vert.y = 24.0f;
        vert.v = 24.0f / 256.0f;
        pvr_prim(&vert, sizeof(pvr_vertex_t));

        vert.x = 560.0f;
        vert.y = 456.0f;
        vert.u = 208.0f / 256.0f;
        vert.v = 168.0f / 256.0f;
        pvr_prim(&vert, sizeof(pvr_vertex_t));

        vert.flags = PVR_CMD_VERTEX_EOL;
        vert.y = 24.0f;
        vert.v = 24.0f / 256.0f;
        pvr_prim(&vert, sizeof(pvr_vertex_t));
    }

    pvr_list_finish();
    pvr_scene_finish();
}

int (*do_frame)(int);

int sms_frame_ntsc(int run) {
    static int16 bufl[NTSC_SAMPLES_PER_FRAME * 2],
        bufr[NTSC_SAMPLES_PER_FRAME * 2];
    static int samples = 0;
    static int frame = 0;
    int line;

    sms_cycles_run = run;
    sms_cycles_to_run = 0;

    for(line = 0; line < NTSC_LINES_PER_FRAME; ++line)  {
        sms_cycles_to_run += SMS_CYCLES_PER_LINE;
        sms_cycles_run += sms_vdp_execute(line);
        sms_cycles_run += sms_z80_run(sms_cycles_to_run - sms_cycles_run);

        sn76489_execute_samples_dc(&psg, bufl + samples, bufr + samples,
                                   psg_samples[line]);
        samples += psg_samples[line];
    }

    frame ^= 1;

    pvr_txr_load_dma(smsvdp.framebuffer, txr, 256 * 256 * 2, 0, dma_done, 0);

    if(frame == 0)  {
        sound_update_buffer_noint(bufl, bufr, NTSC_SAMPLES_PER_FRAME << 2);
        samples = 0;
    }

    return sms_cycles_run - sms_cycles_to_run;
}

int sms_frame_pal(int run)  {
    static int16 bufl[PAL_SAMPLES_PER_FRAME * 2],
        bufr[PAL_SAMPLES_PER_FRAME * 2];
    static int samples = 0;
    static int frame = 0;
    int line;

    sms_cycles_run = run;
    sms_cycles_to_run = 0;

    for(line = 0; line < PAL_LINES_PER_FRAME; ++line)   {
        sms_cycles_to_run += SMS_CYCLES_PER_LINE;
        sms_cycles_run += sms_vdp_execute(line);
        sms_cycles_run += sms_z80_run(sms_cycles_to_run - sms_cycles_run);

        sn76489_execute_samples_dc(&psg, bufl + samples, bufr + samples,
                                   psg_samples[line]);
        samples += psg_samples[line];
    }

    frame ^= 1;

    pvr_txr_load_dma(smsvdp.framebuffer, txr, 256 * 256 * 2, 0, dma_done, 0);

    if(frame == 0)  {
        sound_update_buffer_noint(bufl, bufr, PAL_SAMPLES_PER_FRAME << 2);
        samples = 0;
    }

    return sms_cycles_run - sms_cycles_to_run;
}

void emu_loop(void) {
    maple_device_t *cont1, *cont2;
    cont_state_t *state;
    uint16 mask;
    int oldrun = 0;

#ifdef TIMING_TEST
    int frames = 0;
    time_t thetime = time(NULL);
#endif

    cont1 = maple_enum_type(0, MAPLE_FUNC_CONTROLLER);
    cont2 = maple_enum_type(1, MAPLE_FUNC_CONTROLLER);

    for(;;) {
        oldrun = do_frame(oldrun);
        render_scene();

        mask = 0x0000;

        if(cont1)   {
            state = (cont_state_t *)maple_dev_status(cont1);
            if(state)   {
                mask = ((state->buttons & 0xF0) >> 4) |
                    ((state->buttons & 0x02) << 4) |
                    ((state->buttons & 0x04) << 2);

                if(state->buttons & CONT_START)
                    sms_button_pressed(GG_START);
                else
                    sms_button_released(GG_START);

                if(state->rtrig > 20 && state->buttons & CONT_Y)
                    sms_write_cartram_to_file();

                if(state->ltrig > 20 && state->buttons & CONT_Y)    {
                    sms_write_cartram_to_file();
                    sms_reset();
                    sms_mem_load_rom(menu("/cd/crabemu/roms"));
                }
            }
        }

        if(cont2)   {
            state = (cont_state_t *)maple_dev_status(cont2);
            if(state)   {
                mask |= ((state->buttons & 0xF0) << 2) |
                    ((state->buttons & 0x02) << 10) |
                    ((state->buttons & 0x04) << 8);
            }
        }

        sms_pad = mask ^ 0xFFFF;

#ifdef TIMING_TEST
        ++frames;

        if(thetime + 5 <= time(NULL))   {
            printf("%d frames in %d seconds: %f FPS\n", frames, time(NULL) - thetime,
                   (float) frames / (time(NULL) - thetime));
            frames = 0;
            thetime = time(NULL);
        }
#endif
    }
}

void gui_set_aspect(float x __UNUSED__, float y __UNUSED__) {
    /* XXXX */
}

void gui_set_title(const char *str __UNUSED__)  {
    /* XXXX */
}

void cont_callback(uint8 addr __UNUSED__, uint32 buttons __UNUSED__)    {
    exit(0);
}

int main(int argc __UNUSED__, char *argv[] __UNUSED__)  {
    pvr_poly_cxt_t cxt;
    int dc_region, ct;

    printf("...\n");

    dc_region = flashrom_get_region();
    ct = vid_check_cable();

    /* Prompt the user for whether to run in PAL50 or PAL60 if the flashrom says
       the Dreamcast is European and a VGA Box is not hooked up. */ 
    if(dc_region == FLASHROM_REGION_EUROPE && ct != CT_VGA) {
        if(pal_menu())  {
            vid_set_mode(DM_640x480_NTSC_IL, PM_RGB565);
            region = SMS_VIDEO_NTSC;
            do_frame = &sms_frame_ntsc;
        }
        else    {
            vid_set_mode(DM_640x480_PAL_IL, PM_RGB565);
            region = SMS_VIDEO_PAL;
            do_frame = &sms_frame_pal;
        }
    }
    else    {
        region = SMS_VIDEO_NTSC;
        do_frame = &sms_frame_ntsc;
    }

    pvr_init_defaults();

    dma_sem = sem_create(0);

    cont_btn_callback(0, CONT_A | CONT_B | CONT_X | CONT_Y, cont_callback);

    tex[0] = pvr_mem_malloc(256 * 256 * 2);
    tex[1] = pvr_mem_malloc(256 * 256 * 2);

    pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY,
                     PVR_TXRFMT_RGB565 | PVR_TXRFMT_NONTWIDDLED, 256, 256,
                     tex[0], PVR_FILTER_NONE);
    pvr_poly_compile(&hdr[0], &cxt);

    pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY,
                     PVR_TXRFMT_RGB565 | PVR_TXRFMT_NONTWIDDLED, 256, 256,
                     tex[1], PVR_FILTER_NONE);
    pvr_poly_compile(&hdr[1], &cxt);

    active_tex = 0;
    txr = tex[0];

    sms_mem_load_rom(menu("/cd/crabemu/roms"));

    emu_loop();
    return 0;
}
