/*
    This file is part of CrabEmu.

    Copyright (C) 2007, 2008, 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/queue.h>

#include <kos/fs.h>
#include <png/png.h>
#include <dc/pvr.h>
#include <dc/maple.h>
#include <dc/maple/controller.h>
#include <arch/timer.h>

#include "sms.h"
#include "smsmem.h"
#include "smsmem-gg.h"
#include "font.h"
#include "sound.h"

static char filename[MAX_FN_LEN];
static int firstTime = 1;

struct RomData  {
    dirent_t dirent;
    int type;

    STAILQ_ENTRY(RomData) romlist;
};

STAILQ_HEAD(RomDataList, RomData);
static struct RomDataList romList;
static char **filenames;
static int files = 0;

extern int sms_initialized;
extern int region;

static pvr_ptr_t load_texture(const char *fn, uint32 mask, pvr_list_t list,
                              pvr_poly_hdr_t *hdr)  {
    pvr_poly_cxt_t cxt;
    pvr_ptr_t rv;
    uint32 w, h;
    uint32 fmt;

    if(mask == PNG_NO_ALPHA)    {
        fmt = PVR_TXRFMT_RGB565;
    }
    else if(mask == PNG_MASK_ALPHA) {
        fmt = PVR_TXRFMT_ARGB1555;
    }
    else    {
        fmt = PVR_TXRFMT_ARGB4444;
    }

    png_load_texture(fn, &rv, mask, &w, &h);
    pvr_poly_cxt_txr(&cxt, list, fmt, w, h, rv, PVR_FILTER_NONE);
    pvr_poly_compile(hdr, &cxt);

    return rv;
}

static void create_list(const char *dir)    {
    file_t d;
    dirent_t *ent;
    int len, type, i = 0;
    struct RomData *data;

    STAILQ_INIT(&romList);

    d = fs_open(dir, O_DIR);

    while((ent = fs_readdir(d)))    {
        /* Ignore subdirectories */
        if(ent->attr & O_DIR)   {
            continue;
        }

        len = strlen(ent->name);
        type = -1;

        /* Determine if this rom is of a supported console */
        if(tolower(ent->name[len - 3]) == 's' &&
           tolower(ent->name[len - 2]) == 'm' &&
           tolower(ent->name[len - 1]) == 's')  {
            type = CONSOLE_SMS;
        }
        else if(tolower(ent->name[len - 2]) == 'g' &&
                tolower(ent->name[len - 1]) == 'g') {
            type = CONSOLE_GG;
        }
        else if(tolower(ent->name[len - 2]) == 's' &&
                tolower(ent->name[len - 1]) == 'g') {
            type = CONSOLE_SG1000;
        }

        /* Add the rom to the list if it is supported */
        if(type != -1)  {
            data = (struct RomData *)malloc(sizeof(struct RomData));
            data->type = type;
            data->dirent = *ent;

            STAILQ_INSERT_TAIL(&romList, data, romlist);
            ++files;
        }
    }

    fs_close(d);

    filenames = (char **)malloc(sizeof(char *) * files);

    STAILQ_FOREACH(data, &romList, romlist) {
        filenames[i] = (char *)malloc(strlen(data->dirent.name) + 1);
        strcpy(filenames[i], data->dirent.name);
        ++i;
    }
}

static void destroy_list()  {
    struct RomData *ent, *ent2;
    int i;

    ent = STAILQ_FIRST(&romList);
    while(ent)  {
        ent2 = STAILQ_NEXT(ent, romlist);
        STAILQ_REMOVE_HEAD(&romList, romlist);
        free(ent);
        ent = ent2;
    }

    for(i = 0; i < files; ++i)  {
        free(filenames[i]);
    }

    free(filenames);
    files = 0;
}

static void draw_textured_quad(pvr_poly_hdr_t *hdr, float x, float y, float z,
                               float w, float h, float umax, float vmax)    {
    pvr_vertex_t vert;

    pvr_prim(hdr, sizeof(pvr_poly_hdr_t));

    vert.flags = PVR_CMD_VERTEX;
    vert.argb = 0xFFFFFFFF;
    vert.oargb = 0x00000000;

    vert.x = x;
    vert.y = y + h;
    vert.z = z;
    vert.u = 0.0f;
    vert.v = vmax;
    pvr_prim(&vert, sizeof(pvr_vertex_t));

    vert.y = y;
    vert.v = 0.0f;
    pvr_prim(&vert, sizeof(pvr_vertex_t));

    vert.x = x + w;
    vert.y = y + h;
    vert.u = umax;
    vert.v = vmax;
    pvr_prim(&vert, sizeof(pvr_vertex_t));

    vert.flags = PVR_CMD_VERTEX_EOL;
    vert.y = y;
    vert.v = 0.0f;
    pvr_prim(&vert, sizeof(pvr_vertex_t));
}

static void load_bios_sms(void) {
    FILE *fp;

    fp = fopen("/cd/crabemu/bios.sms", "rb");
    if(fp)  {
        fclose(fp);
        sms_mem_load_bios("/cd/crabemu/bios.sms");
        sms_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                               SMS_MEMCTL_RAM) ^ 0xFF);
    }
}

static void load_bios_gg(void)  {
    FILE *fp;

    fp = fopen("/cd/crabemu/bios.gg", "rb");
    if(fp)  {
        fclose(fp);
        sms_mem_load_bios("/cd/crabemu/bios.gg");
        sms_gg_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                                 SMS_MEMCTL_CART | SMS_MEMCTL_RAM) ^ 0xFF);
    }
}

const char *menu(const char *dir)   {
    pvr_ptr_t bg, selector, sms, gg, sg1000, font;
    pvr_poly_hdr_t hdr_bg, hdr_selector, hdr_sms, hdr_gg, hdr_sg1000, hdr_title;
    int topItem = 0;
    int selected = 0;
    int selectedType = -1;
    int i;
    maple_device_t *cont1;
    cont_state_t *state;
    uint64 timer = timer_ms_gettime64();

    if(firstTime)   {
        pvr_ptr_t title;

        title = load_texture("/cd/crabemu/title.png", PNG_NO_ALPHA,
                             PVR_LIST_OP_POLY, &hdr_title);

        timer = timer_ms_gettime64();

        while(timer + 5000 > timer_ms_gettime64())  {
            pvr_wait_ready();
            pvr_scene_begin();
            
            pvr_list_begin(PVR_LIST_OP_POLY);
            draw_textured_quad(&hdr_title, 0.0f, 0.0f, 1.0f, 640.0f, 480.0f,
                               640.0f / 1024.0f, 480.0f / 512.0f);
            
            pvr_list_finish();
            pvr_scene_finish();
        }

        firstTime = 0;
        pvr_mem_free(title);
    }

    bg = load_texture("/cd/crabemu/menubg.png", PNG_NO_ALPHA, PVR_LIST_OP_POLY,
                      &hdr_bg);
    selector = load_texture("/cd/crabemu/selector.png", PNG_FULL_ALPHA,
                            PVR_LIST_TR_POLY, &hdr_selector);
    sms = load_texture("/cd/crabemu/sms.png", PNG_NO_ALPHA, PVR_LIST_OP_POLY,
                       &hdr_sms);
    gg = load_texture("/cd/crabemu/gg.png", PNG_NO_ALPHA, PVR_LIST_OP_POLY,
                      &hdr_gg);
    sg1000 = load_texture("/cd/crabemu/sg1000.png", PNG_NO_ALPHA,
                          PVR_LIST_OP_POLY, &hdr_sg1000);
    font = font_load_txr();

    cont1 = maple_enum_type(0, MAPLE_FUNC_CONTROLLER);

    while(files == 0)   {
        create_list(dir);

        pvr_wait_ready();
        pvr_scene_begin();

        pvr_list_begin(PVR_LIST_OP_POLY);
        draw_textured_quad(&hdr_bg, 0.0f, 0.0f, 1.0f, 640.0f, 480.0f,
                           640.0f / 1024.0f, 480.0f / 512.0f);

        pvr_list_finish();

        pvr_list_begin(PVR_LIST_TR_POLY);
        font_draw_str("Please insert a ROM CD.", font, 250, 70, 2);
        pvr_list_finish();
        pvr_scene_finish();
    }

    for(;;) {
        if(timer + 200 < timer_ms_gettime64())  {
            state = (cont_state_t *)maple_dev_status(cont1);

            if(state)   {
                if(state->buttons & CONT_DPAD_DOWN) {
                    ++selected;

                    if(selected == files)   {
                        selected = files - 1;
                    }
                    else if(selected == 13) {
                        if(topItem + 13 < files)    {
                            ++topItem;
                        }

                        selected = 12;
                    }

                    selectedType = -1;
                    timer = timer_ms_gettime64();
                }
                else if(state->buttons & CONT_DPAD_UP)  {
                    --selected;

                    if(selected == -1)  {
                        if(topItem) {
                            --topItem;
                        }

                        selected = 0;
                    }

                    selectedType = -1;
                    timer = timer_ms_gettime64();
                }

                if(state->buttons & CONT_DPAD_RIGHT)    {
                    if(topItem + 26 <= files)   {
                        topItem += 13;
                    }
                    else    {
                        selected = 12;
                        topItem = files - 13;
                    }

                    selectedType = -1;
                    timer = timer_ms_gettime64();
                }
                else if(state->buttons & CONT_DPAD_LEFT)    {
                    if(topItem < 13)    {
                        topItem = 0;
                        selected = 0;
                    }
                    else    {
                        topItem -= 13;
                    }

                    selectedType = -1;
                    timer = timer_ms_gettime64();
                }

                if(state->buttons & CONT_A) {
                    sprintf(filename, "%s/%s", dir,
                            filenames[selected + topItem]);
                    break;
                }
            }
        }
    
        if(selectedType == -1)  {
            char *str = filenames[selected + topItem];
            int len = strlen(str);

            if(tolower(str[len - 3]) == 's' &&
               tolower(str[len - 2]) == 'm' &&
               tolower(str[len - 1]) == 's')    {
                selectedType = CONSOLE_SMS;
            }
            else if(tolower(str[len - 2]) == 'g' &&
                    tolower(str[len - 1]) == 'g')   {
                selectedType = CONSOLE_GG;
            }
            else if(tolower(str[len - 2]) == 's' &&
                    tolower(str[len - 1]) == 'g')   {
                selectedType = CONSOLE_SG1000;
            }
        }

        pvr_wait_ready();
        pvr_scene_begin();

        pvr_list_begin(PVR_LIST_OP_POLY);
        draw_textured_quad(&hdr_bg, 0.0f, 0.0f, 1.0f, 640.0f, 480.0f,
                           640.0f / 1024.0f, 480.0f / 512.0f);

        /* Draw the appropriate icons here (50, 229) - 154x76 */
        if(selectedType == CONSOLE_SMS) {
            draw_textured_quad(&hdr_sms, 50.0f, 229.0f, 1.1f, 154.0f, 76.0f,
                               154.0f / 256.0f, 76.0f / 128.0f);
        }
        else if(selectedType == CONSOLE_GG) {
            draw_textured_quad(&hdr_gg, 50.0f, 229.0f, 1.1f, 154.0f, 76.0f,
                               154.0f / 256.0f, 76.0f / 128.0f);
        }
        else if(selectedType == CONSOLE_SG1000) {
            draw_textured_quad(&hdr_sg1000, 50.0f, 229.0f, 1.1f, 154.0f, 76.0f,
                               154.0f / 256.0f, 76.0f / 128.0f);
        }

        pvr_list_finish();

        pvr_list_begin(PVR_LIST_TR_POLY);
        draw_textured_quad(&hdr_selector, 226.0f, 70.0f + 26 * selected, 1.1f,
                           375.0f, 26.0f, 375.0f / 512.0f, 26.0f / 32.0f);

        for(i = 0; i < 13 && i + topItem < files; ++i)  {
            font_draw_str(filenames[i + topItem], font, 250, 70 + 26 * i, 2);
        }

        pvr_list_finish();
        pvr_scene_finish();
    }
        
    destroy_list();

    pvr_mem_free(bg);
    pvr_mem_free(selector);
    pvr_mem_free(sms);
    pvr_mem_free(gg);
    pvr_mem_free(sg1000);
    pvr_mem_free(font);

    if(!sms_initialized)    {
        sms_init(region, SMS_REGION_EXPORT);
    }
    else    {
        sms_reset();
    }

    if(selectedType == CONSOLE_SMS)
        load_bios_sms();
    else if(selectedType == CONSOLE_GG)
        load_bios_gg();

    return filename;
}
