/*
    This file is part of CrabEmu.

    Copyright (C) 2007, 2008 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* The sound code found herein is roughly based upon code from Genesis Plus/DC,
   which was ported to the Dreamcast by BlackAura (with subsequent modifications
   by Quzar). The sound stuff in GPDC though is pretty much the same as the
   snd_stream stuff in KOS, so I guess Dan Potter should get some credit for
   this code, eh? */

#include <string.h>
#include <dc/spu.h>
#include <dc/g2bus.h>
#include <dc/flashrom.h>
#include <dc/sound/sound.h>

#include "sound.h"
#include "sms.h"
#include <arm/aica_cmd_iface.h>

static const int NTSC_SAMPLES_PER_FRAME = 735;
static const int PAL_SAMPLES_PER_FRAME = 882;

static const int FRAMES_TO_BUFFER = 4;
static const int SAMPLE_RATE = 44100;

static const int POS_OFFSET = 40;
static const int AICA_REGS = 0xa0800000;

static int initted = 0;
static uint32 ptr[2];
static int buf_pos;
static int samples_per_frame;

extern int region;

/* Pretty much the same as dcsnd_stop() in GPDC */
static void sound_stop(void)    {
    AICA_CMDSTR_CHANNEL(tmp, cmd, chan);

    if(!initted)
        return;

    /* Reset the left and right channels */
    cmd->cmd = AICA_CMD_CHAN;
    cmd->timestamp = 0;
    cmd->size = AICA_CMDSTR_CHANNEL_SIZE;
    cmd->cmd_id = 0;
    chan->cmd = AICA_CH_CMD_STOP;
    snd_sh4_to_aica(tmp, cmd->size);

    cmd->cmd_id = 1;
    snd_sh4_to_aica(tmp, cmd->size);

    /* Reset the sound buffers */
    spu_memset(ptr[0], 0, NTSC_SAMPLES_PER_FRAME * FRAMES_TO_BUFFER * 2);
    spu_memset(ptr[1], 0, NTSC_SAMPLES_PER_FRAME * FRAMES_TO_BUFFER * 2);
}

/* This function shamelessly mostly copied from GPDC's dcsnd_play() function.
   Note though, the special handling for a system set to mono mode is gone.
   No Master System games can do stereo sound, and very few Game Gear games
   really seem to take advantage of it (probably because the system itself
   only had one speaker). */
static void sound_start(void)   {
    AICA_CMDSTR_CHANNEL(tmp, cmd, chan);

    if(!initted)
        return;

    /* Make sure these are sync'd (and/or delayed) */
    snd_sh4_to_aica_stop();

    /* Set up channel 0 */
    cmd->cmd = AICA_CMD_CHAN;
    cmd->timestamp = 0;
    cmd->size = AICA_CMDSTR_CHANNEL_SIZE;
    cmd->cmd_id = 0;
    chan->cmd = AICA_CH_CMD_START | AICA_CH_START_DELAY;
    chan->base = ptr[0];
    chan->type = AICA_SM_16BIT;
    chan->length = samples_per_frame * FRAMES_TO_BUFFER;
    chan->loop = 1;
    chan->loopstart = 0;
    chan->loopend = samples_per_frame * FRAMES_TO_BUFFER;
    chan->freq = SAMPLE_RATE;
    chan->vol = 255;
    chan->pan = 0;
    snd_sh4_to_aica(tmp, cmd->size);

    /* Set up channel 1 */
    cmd->cmd_id = 1;
    chan->base = ptr[1];
    chan->pan = 255;
    snd_sh4_to_aica(tmp, cmd->size);

    /* Send synchronised start command */
    cmd->cmd_id = (1 << 0) | (1 << 1);

    chan->cmd = AICA_CH_CMD_START | AICA_CH_START_SYNC;
    snd_sh4_to_aica(tmp, cmd->size);

    /* Set up streaming state */
    buf_pos = 0;

    /* Start both sound buffers playing */
    snd_sh4_to_aica_start();
}

void sound_update_buffer_noint(int16 *bufl, int16 *bufr,
                               int length)  {
    /* UGLY UGLY UGLY UGLY!
       This little hack is for making things play nice, and not go too fast.
       For some reason or another, randomly KOS decides to let me go > 60fps
       (or 50fps, whichever is appropriate), so this makes it so that the sound
       actually becomes the limiting factor... sorta. */
    int pos = g2_read_32(AICA_REGS + AICA_CHANNEL(0) + POS_OFFSET) << 1;

    while(buf_pos < pos && (buf_pos + length) > pos)    {
        pos = g2_read_32(AICA_REGS + AICA_CHANNEL(0) + POS_OFFSET) << 1;
    }

    spu_memload(ptr[0] + buf_pos, bufl, length);
    spu_memload(ptr[1] + buf_pos, bufr, length);

    buf_pos += length;

    if(buf_pos == samples_per_frame * FRAMES_TO_BUFFER * 2)
        buf_pos = 0;
}

int sound_init(void)    {
    if(initted)
        return 0;

    snd_init();

    if(region == SMS_VIDEO_NTSC)
        samples_per_frame = NTSC_SAMPLES_PER_FRAME;
    else
        samples_per_frame = PAL_SAMPLES_PER_FRAME;

    /* Allocate the SPU RAM buffers and clear them out */
    ptr[0] = snd_mem_malloc(samples_per_frame * FRAMES_TO_BUFFER * 2);
    ptr[1] = snd_mem_malloc(samples_per_frame * FRAMES_TO_BUFFER * 2);

    spu_memset(ptr[0], 0x00, samples_per_frame * FRAMES_TO_BUFFER * 2);
    spu_memset(ptr[1], 0x00, samples_per_frame * FRAMES_TO_BUFFER * 2);

    initted = 1;

    sound_start();

    return 0;
}

void sound_shutdown(void)   {
    sound_stop();

    snd_mem_free(ptr[1]);
    snd_mem_free(ptr[0]);

    initted = 0;
}

void sound_reset_buffer(void)   {
    sound_stop();
    spu_memset(ptr[0], 0x00, samples_per_frame * FRAMES_TO_BUFFER * 2);
    spu_memset(ptr[1], 0x00, samples_per_frame * FRAMES_TO_BUFFER * 2);
    sound_start();
}
