/*
	This file is part of CrabEmu.

	Copyright (C) 2005, 2006, 2007, 2008, 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

	CrabEmu is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CrabEmu; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <unistd.h>

#include <OpenGL/OpenGL.h>

#include "EmuController.h"
#include "EmuOGLView.h"
#include "sms.h"
#include "smsmem.h"
#include "smsmem-gg.h"
#include "smsz80.h"
#include "sn76489.h"
#include "sound.h"
#include "smsvdp.h"
#include "joystick.h"

/* From main.m */
extern NSLock *memoryLock;
extern EmuOGLView *theview;
extern void emu_loop();
extern BOOL debuggerEnabled;
extern MPSemaphoreID ss_sem;
extern BOOL ntscVideo;

/* from sms.c */
extern int sms_psg_enabled;
extern sn76489_t psg;
extern int sms_initialized;
extern int sms_console;
extern sms_vdp_t smsvdp;

int started = 0;

@implementation EmuController

- (void)application:(NSApplication *)app openFiles:(NSArray *)filenames
{
    NSString *fn;
    NSDocumentController *c;

    /* Eventually we should do something different if for some reason someone
       tries to send us multiple files, but for now, punt and open the first
       file in the list */
    fn = [filenames objectAtIndex:0];

    [memoryLock lock];

    if([self initMasterSystem] == NO)   {
        printf("could not init master system, bailing out!\n");
        exit(-1);
    }

    if([self loadMasterSystemRomFromFile:fn] == NO) {
        NSString *title = [NSString stringWithFormat:@"Cannot open ROM"
            " file %@.", fn];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read from"
                        "this directory, and this file.", 
                        @"OK", NULL, NULL);

        [memoryLock unlock];
        
        [NSApp replyToOpenOrPrint:NSApplicationDelegateReplyFailure];
        return;
    }

    c = [NSDocumentController sharedDocumentController];
    [c noteNewRecentDocumentURL:[NSURL fileURLWithPath:fn]];

    [memoryLock unlock];

    [NSApp replyToOpenOrPrint:NSApplicationDelegateReplySuccess];
}

-(NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)app
{
    if(sms_initialized == 1)
        sms_write_cartram_to_file();

    if(j1map != nil)    {
        [j1map release];
    }

    if(j2map != nil)    {
        [j2map release];
    }

    /* Release all joysticks. This will close any open joysticks as well. */
    joy_release_joysticks();

    return NSTerminateNow;
}

- (void)awakeFromNib
{
    BOOL palMode, smsBiosOn, ggBiosOn, disableSaver;
    unichar string[14];
    NSMutableDictionary *keys;
    int i;

    /* Make the pattern and palette panels not able to become key windows for
       the application. This fixes some potential problems with them and
       fullscreen mode. */
    [vdpPalettePanel setBecomesKeyOnlyIfNeeded:YES];
    [vdpPatternPanel setBecomesKeyOnlyIfNeeded:YES];

    prefs = [[NSUserDefaults standardUserDefaults] retain];

    openRom = nil;
    started = 0;

    palMode = [prefs boolForKey:@"SMS PAL"];
    domestic = [prefs boolForKey:@"SMS Domestic"];
    smsBiosOn = [prefs boolForKey:@"SMS BIOS Enabled"];
    ggBiosOn = [prefs boolForKey:@"GG BIOS Enabled"];
    disableSaver = [prefs boolForKey:@"Disable Screensaver"];

    if([prefs dictionaryForKey:@"SMS Buttonmap"] != nil)    {
        /* Use CrabEmu 0.1.9 and later style preferences. */
        keymap =
            (NSMutableDictionary *)[prefs dictionaryForKey:@"SMS Buttonmap"];
        [keymap retain];

        keys = [[keymap objectForKey:@"Keyboard 1"] retain];
        [self fillString:string withDictionary:keys];
        [keys release];

        keys = [[keymap objectForKey:@"Keyboard 2"] retain];
        [self fillString:string + 7 withDictionary:keys];
        [keys release];

        buttonmap = [[NSString stringWithCharacters:string length:14] retain];
    }
    else if([prefs stringForKey:@"SMS Buttons"] != nil) {
        /* Update CrabEmu 0.1.8 and earlier style preferences. */
        keymap = [[NSMutableDictionary dictionaryWithCapacity:2] retain];
        buttonmap = [prefs stringForKey:@"SMS Buttons"];
        [buttonmap retain];
        [self savePrefs];
    }
    else    {
        /* Create new CrabEmu 0.1.9 and later style preferences. */
        string[0] = 0x7E;
        string[1] = 0x7D;
        string[2] = 0x7B;
        string[3] = 0x7C;
        string[4] = 0x29;
        string[5] = 0x27;
        string[6] = 0x23;
        string[7] = 0x0D;
        string[8] = 0x01;
        string[9] = 0x00;
        string[10] = 0x02;
        string[11] = 0x09;
        string[12] = 0x0B;
        string[13] = 0x69;

        keymap = [[NSMutableDictionary dictionaryWithCapacity:2] retain];
        buttonmap = [[NSString stringWithCharacters:string length:14] retain];
        [self savePrefs];
    }

    if([prefs stringForKey:@"SMS BIOS ROM"] != nil) {
        [smsBios setStringValue:[prefs stringForKey:@"SMS BIOS ROM"]];
    }

    if([prefs stringForKey:@"Game Gear BIOS ROM"] != nil)   {
        [ggBios setStringValue:[prefs stringForKey:@"Game Gear BIOS ROM"]];
    }

    if(palMode) {
        [pal setState:NSOnState];
        [ntsc setState:NSOffState];
        ntscVideo = NO;
    }

    if(domestic)    {
        [rDomestic setState:NSOnState];
        [rExport setState:NSOffState];
        domestic = YES;
    }

    if(smsBiosOn)   {
        [smsBiosEnabled setState:NSOnState];
    }
    else    {
        [smsBiosEnabled setState:NSOffState];
    }

    if(ggBiosOn)    {
        [ggBiosEnabled setState:NSOnState];
    }
    else    {
        [ggBiosEnabled setState:NSOffState];
    }

    if(disableSaver)    {
        [ssaverDisabled setState:NSOnState];
        timer = [NSTimer scheduledTimerWithTimeInterval:15
                                                 target:self
                                               selector:@selector(ssTimer:)
                                               userInfo:nil
                                                repeats:YES];
    }
    else    {
        [ssaverDisabled setState:NSOffState];
        timer = nil;
    }

    if(CGLLockContext == NULL)  {
        [fullscreen setEnabled:NO];
    }

    j1map = nil;
    j2map = nil;
    joy1 = NULL;
    joy2 = NULL;

    p1curinput = p1keyboard;
    p2curinput = p2keyboard;

    [screenshot setEnabled:NO];
    [fullscreen setEnabled:NO];
    [pause setEnabled:NO];
    [reset setEnabled:NO];

    for(i = 0; i < [saveStates numberOfItems]; ++i) {
        [[saveStates itemAtIndex:i] setEnabled:NO];
    }

    for(i = 0; i < [loadStates numberOfItems]; ++i) {
        [[loadStates itemAtIndex:i] setEnabled:NO];
    }

    [saveSRAM setEnabled:NO];
    [self scanJoysticks:nil];
}

- (void)dealloc
{
    [openRom release];
    [prefs release];
    [buttonmap release];
    [keymap release];

    [super dealloc];
}

- (IBAction)enableNoise:(id)sender
{
	if([noise state] == NSOnState)	{
		[noise setState: NSOffState];
		psg.enabled_channels &= ~ENABLE_NOISE;
	}
	else	{
		[noise setState: NSOnState];
		psg.enabled_channels |= ENABLE_NOISE;
	}
}

- (IBAction)enablePSG:(id)sender
{
	if([PSG state] == NSOnState)	{
		[PSG setState:NSOffState];
		sms_psg_enabled = 0;
	}
	else	{
		[PSG setState:NSOnState];
		sms_psg_enabled = 1;
	}
}

- (IBAction)enableTone0:(id)sender
{
	if([tone0 state] == NSOnState)	{
		[tone0 setState:NSOffState];
		psg.enabled_channels &= ~ENABLE_TONE0;
	}
	else	{
		[tone0 setState:NSOnState];
		psg.enabled_channels |= ENABLE_TONE0;
	}
}

- (IBAction)enableTone1:(id)sender
{
	if([tone1 state] == NSOnState)	{
		[tone1 setState:NSOffState];
		psg.enabled_channels &= ~ENABLE_TONE1;
	}
	else	{
		[tone1 setState:NSOnState];
		psg.enabled_channels |= ENABLE_TONE1;
	}
}

- (IBAction)enableTone2:(id)sender
{
	if([tone2 state] == NSOnState)	{
		[tone2 setState:NSOffState];
		psg.enabled_channels &= ~ENABLE_TONE2;
	}
	else	{
		[tone2 setState:NSOnState];
		psg.enabled_channels |= ENABLE_TONE2;
	}
}

- (IBAction)loadMasterRom:(id)sender
{
	NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    NSArray *fileTypes = [NSArray arrayWithObjects:@"sms", @"gg", @"sc", @"sg", nil];
	[openPanel setTitle:@"Choose a Master System/Game Gear Rom"];
	[openPanel setCanChooseDirectories:NO];
	
	if([openPanel runModalForTypes:fileTypes] == NSOKButton)	{
		NSString *fn = [openPanel filename];
        NSDocumentController *c;

        [memoryLock lock];

        if([self initMasterSystem] == NO)   {
            printf("could not init master system, bailing out!\n");
            exit(-1);
        }

        if([self loadMasterSystemRomFromFile:fn] == NO) {
            NSString *title = [NSString stringWithFormat:@"Cannot open ROM"
                " file %@.", fn];
            NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                            "Verify that you have the permissions to read from"
                            "this directory, and this file.", 
                            @"OK", NULL, NULL);
        }

        c = [NSDocumentController sharedDocumentController];
        [c noteNewRecentDocumentURL:[NSURL fileURLWithPath:fn]];

        [memoryLock unlock];
	}
}

- (IBAction)prefGGStart:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[6] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (IBAction)prefSms1B1:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];

    buttons[4] = [self runVirtKeyboard];

    [buttonmap release];

    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (IBAction)prefSms1B2:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[5] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (IBAction)prefSms1Down:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[1] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (IBAction)prefSms1Left:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[2] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (IBAction)prefSms1Reset:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[0] = 0x7E;
    buttons[1] = 0x7D;
    buttons[2] = 0x7B;
    buttons[3] = 0x7C;
    buttons[4] = 0x29;
    buttons[5] = 0x27;
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}    

- (IBAction)prefSms1Right:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[3] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (IBAction)prefSms1Up:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[0] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (IBAction)prefSms2B1:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[11] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (IBAction)prefSms2B2:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[12] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (IBAction)prefSms2Down:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[8] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (IBAction)prefSms2Left:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[9] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (IBAction)prefSms2Reset:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[7] = 0x0D;
    buttons[8] = 0x01;
    buttons[9] = 0x00;
    buttons[10] = 0x02;
    buttons[11] = 0x09;
    buttons[12] = 0x0B;
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (IBAction)prefSms2Right:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[10] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (IBAction)prefSms2Up:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[7] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];

    [self savePrefs];
}

- (void)handleSMSButton:(int)button withJoystickElement:(NSNumber *)n
{
    uint32 num = [n unsignedLongValue];

    if(num & 0x80000000)    {
        /* Hat switch */
        if(((joy_read_hat(joy1, num & 0xFF) << 8) &
            (int)(num & 0xFF00)) == (int)(num & 0xFF00))
            sms_button_pressed(button);
        else
            sms_button_released(button);
    }
    else if(num & 0x40000000)   {
        /* Axis */
        if(num & 0x20000000)    {
            if(joy_read_axis(joy1, num & 0xFF) < -16384)
                sms_button_pressed(button);
            else
                sms_button_released(button);
        }
        else    {
            if(joy_read_axis(joy1, num & 0xFF) > 16384)
                sms_button_pressed(button);
            else
                sms_button_released(button);
        }
    }
    else    {
        /* Button */
        if(joy_read_button(joy1, (int)num) >= 1)
            sms_button_pressed(button);
        else
            sms_button_released(button);
    }
}

- (void)handleJoysticks
{
    NSNumber *n;

    if(j1map)   {
        n = (NSNumber *)[j1map objectForKey:@"SMS Up"];
        [self handleSMSButton:SMS_PAD1_UP withJoystickElement:n];

        n = (NSNumber *)[j1map objectForKey:@"SMS Down"];
        [self handleSMSButton:SMS_PAD1_DOWN withJoystickElement:n];

        n = (NSNumber *)[j1map objectForKey:@"SMS Left"];
        [self handleSMSButton:SMS_PAD1_LEFT withJoystickElement:n];

        n = (NSNumber *)[j1map objectForKey:@"SMS Right"];
        [self handleSMSButton:SMS_PAD1_RIGHT withJoystickElement:n];

        n = (NSNumber *)[j1map objectForKey:@"SMS Button 1"];
        [self handleSMSButton:SMS_PAD1_A withJoystickElement:n];

        n = (NSNumber *)[j1map objectForKey:@"SMS Button 2"];
        [self handleSMSButton:SMS_PAD1_B withJoystickElement:n];

        n = (NSNumber *)[j1map objectForKey:@"Game Gear Start"];
        [self handleSMSButton:GG_START withJoystickElement:n];
    }

    if(j2map)   {
        n = (NSNumber *)[j2map objectForKey:@"SMS Up"];
        [self handleSMSButton:SMS_PAD2_UP withJoystickElement:n];

        n = (NSNumber *)[j2map objectForKey:@"SMS Down"];
        [self handleSMSButton:SMS_PAD2_DOWN withJoystickElement:n];

        n = (NSNumber *)[j2map objectForKey:@"SMS Left"];
        [self handleSMSButton:SMS_PAD2_LEFT withJoystickElement:n];

        n = (NSNumber *)[j2map objectForKey:@"SMS Right"];
        [self handleSMSButton:SMS_PAD2_RIGHT withJoystickElement:n];

        n = (NSNumber *)[j2map objectForKey:@"SMS Button 1"];
        [self handleSMSButton:SMS_PAD2_A withJoystickElement:n];

        n = (NSNumber *)[j2map objectForKey:@"SMS Button 2"];
        [self handleSMSButton:SMS_PAD2_B withJoystickElement:n];
    }
}

- (void)updatePalette
{
    if([showVDPPalette state] == NSOnState) {
        [paletteView setNeedsDisplay:YES];
    }
}

- (void)updatePatterns
{
    if([showVDPPatterns state] == NSOnState) {
        [patternView setNeedsDisplay:YES];
    }
}

- (unichar)runVirtKeyboard
{
    unichar retval;

    [NSApp beginSheet:virtKeyboard
       modalForWindow:prefWindow
        modalDelegate:self
       didEndSelector:NULL
          contextInfo:NULL];

    retval = (unichar)[NSApp runModalForWindow:virtKeyboard];

    [NSApp endSheet:virtKeyboard];
    [virtKeyboard orderOut:nil];

    return retval;
}

- (void)endVirtKeyboard:(unichar)character
{
    [NSApp stopModalWithCode:(int)character];
}

- (void)masterEmuThread:(id)junk
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

	emu_loop();

	sms_shutdown();
	[pool release];
}

- (void)pauseEmulation:(id)sender
{
    if(started == 0)
        return;

    if(sms_console == CONSOLE_SMS)  {
        [memoryLock lock];
        sms_z80_nmi();
        [memoryLock unlock];
    }

    if([pause state] == NSOffState) {
        if(sms_console == CONSOLE_GG)   {
            [memoryLock lock];
            sound_pause();
        }

        [pause setState: NSOnState];
        
    }
    else    {
        if(sms_console == CONSOLE_GG)   {
            sound_unpause();
            [memoryLock unlock];
        }

        [pause setState: NSOffState];
    }
}

- (IBAction)enableDebugger:(id)sender
{
    if([debugger state] == NSOffState) {
        [debugger setState:NSOnState];
        [smsVdpDebuggerWindow makeKeyAndOrderFront:self];
        [debuggerWindow makeKeyAndOrderFront:self];
        debuggerEnabled = YES;
    }
    else    {
        [debugger setState:NSOffState];
        [smsVdpDebuggerWindow orderOut:self];
        [debuggerWindow orderOut:self];
        debuggerEnabled = NO;
        MPSignalSemaphore(ss_sem);
    }
}

- (NSString *)buttonMap
{
    unichar str[14];

    /* Disable the keyboard for any players that are using joysticks. */
    if([p1keyboard state] == NSOnState && [p2keyboard state] == NSOnState)  {
        return buttonmap;
    }
    else if([p1keyboard state] == NSOffState &&
            [p2keyboard state] == NSOffState)   {
        memset(str, 0xFF, 14 * sizeof(unichar));
        return [NSString stringWithCharacters:str length:14];
    }
    else if([p1keyboard state] == NSOnState)    {
        [buttonmap getCharacters:str];
        memset(str + 7, 0xFF, 7 * sizeof(unichar));
        return [NSString stringWithCharacters:str length:14];
    }
    else    {
        [buttonmap getCharacters:str];
        memset(str, 0xFF, 7 * sizeof(unichar));
        return [NSString stringWithCharacters:str length:14];
    }
}

- (IBAction)saveSRAM:(id)sender
{
    if(sms_initialized == 1)
        sms_write_cartram_to_file();
}

- (IBAction)saveState1:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".1.cestate"];

    [memoryLock lock];

    if(sms_save_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to this "
                        "directory and that the target volume is not full.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)saveState2:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".2.cestate"];

    [memoryLock lock];

    if(sms_save_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to this "
                        "directory and that the target volume is not full.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)saveState3:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".3.cestate"];

    [memoryLock lock];

    if(sms_save_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to this "
                        "directory and that the target volume is not full.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)saveState4:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".4.cestate"];

    [memoryLock lock];

    if(sms_save_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to this "
                        "directory and that the target volume is not full.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)saveState5:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".5.cestate"];

    [memoryLock lock];

    if(sms_save_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to this "
                        "directory and that the target volume is not full.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)saveState6:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".6.cestate"];

    [memoryLock lock];

    if(sms_save_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to this "
                        "directory and that the target volume is not full.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)loadState1:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".1.cestate"];

    [memoryLock lock];

    if(sms_load_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read to this "
                        "directory and that the file exists.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)loadState2:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".2.cestate"];

    [memoryLock lock];

    if(sms_load_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read to this "
                        "directory and that the file exists.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)loadState3:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".3.cestate"];

    [memoryLock lock];

    if(sms_load_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read to this "
                        "directory and that the file exists.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)loadState4:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".4.cestate"];

    [memoryLock lock];

    if(sms_load_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read to this "
                        "directory and that the file exists.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)loadState5:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".5.cestate"];

    [memoryLock lock];

    if(sms_load_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read to this "
                        "directory and that the file exists.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)loadState6:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".6.cestate"];

    [memoryLock lock];

    if(sms_load_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read to this "
                        "directory and that the file exists.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)resetEmulation:(id)sender
{
    sms_soft_reset();
}

- (IBAction)modeNTSC:(id)sender
{
    if(sms_initialized != 1)    {
        if([ntsc state] == NSOffState) {
            [ntsc setState:NSOnState];
            [pal setState:NSOffState];
            ntscVideo = YES;
        }

        [self savePrefs];
    }
}

- (IBAction)modePAL:(id)sender
{
    if(sms_initialized != 1)    {
        if([pal state] == NSOffState) {
            [pal setState:NSOnState];
            [ntsc setState:NSOffState];
            ntscVideo = NO;
        }

        [self savePrefs];
    }
}

- (IBAction)regionDomestic:(id)sender
{
    if(sms_initialized != 1)    {
        if([rDomestic state] == NSOffState) {
            [rDomestic setState:NSOnState];
            [rExport setState:NSOffState];
            domestic = YES;
        }
        
        [self savePrefs];
    }
}

- (IBAction)regionExport:(id)sender
{
    if(sms_initialized != 1)    {
        if([rExport state] == NSOffState) {
            [rExport setState:NSOnState];
            [rDomestic setState:NSOffState];
            domestic = NO;
        }
        
        [self savePrefs];
    }
}

- (IBAction)scanJoysticks:(id)sender
{
    joydata_t *joy;
    NSString *str;
    NSMenuItem *item, *item2;
    int i;

    /* Release any already scanned in joysticks. */
    joy_release_joysticks();

    while([p1input numberOfItems] > 1)  {
        [p1input removeItemAtIndex:1];
        [p2input removeItemAtIndex:1];
    }

    [pjoysticks removeAllItems];

    /* Scan for any joysticks attached to the system. */
    joy_count = joy_scan_joysticks();

    if(!joy_count)  {
        [pjoysticks addItemWithTitle:@"No Joysticks Detected"];
        [pjoysticks setEnabled:NO];
        [pjoyUp setEnabled:NO];
        [pjoyRight setEnabled:NO];
        [pjoyDown setEnabled:NO];
        [pjoyLeft setEnabled:NO];
        [pjoyB1 setEnabled:NO];
        [pjoyB2 setEnabled:NO];
        [pjoyGGStart setEnabled:NO];
    }
    else    {
        [pjoysticks setEnabled:YES];
        [pjoyUp setEnabled:YES];
        [pjoyRight setEnabled:YES];
        [pjoyDown setEnabled:YES];
        [pjoyLeft setEnabled:YES];
        [pjoyB1 setEnabled:YES];
        [pjoyB2 setEnabled:YES];
        [pjoyGGStart setEnabled:YES];
    }

    for(i = 0; i < joy_count; ++i)  {
        joy = joy_get_joystick(i);
        str = [NSString stringWithUTF8String:joy->name];

        item = [[NSMenuItem alloc] initWithTitle:str
                                          action:@selector(setInput1:)
                                   keyEquivalent:@""];

        item2 = [[NSMenuItem alloc] initWithTitle:str
                                           action:@selector(setInput2:)
                                    keyEquivalent:@""];

        /* Set the tag on the menu items. */
        [item setTag:i];
        [item2 setTag:i];

        /* Set the target of the menu item to this object. */
        [item setTarget:self];
        [item2 setTarget:self];

        /* Add the items to the menus. */
        [p1input addItem:item];
        [p2input addItem:item2];

        /* Check if the device is configured. */
        if([keymap objectForKey:str] == nil)    {
            [item setEnabled:NO];
            [item2 setEnabled:NO];
        }
        else    {
            NSDictionary *d = (NSDictionary *)[keymap objectForKey:str];

            if([d count] < 7)   {
                [item setEnabled:NO];
                [item2 setEnabled:NO];
            }
        }

        [item release];
        [item2 release];

        [pjoysticks addItemWithTitle:str];
    }
}

- (IBAction)setInput1:(id)sender
{
    NSMenuItem *item = (NSMenuItem *)sender;
    int tag;
    NSString *str;
    joydata_t *old = joy1;

    /* If we have tried to select the same input device we currently are using,
       ignore it. */
    if([item state] == NSOnState)   {
        return;
    }

    /* Lock the memory lock, and grab the tag of the menu item. */
    [memoryLock lock];
    tag = [item tag];

    /* If the user has selected the keyboard, close the old joytick if any,
       and release the old button mapping. */
    if(tag == -1)   {
        if(old) {
            /* Release resources associated with the joystick. */
            joy_close_joystick(old);
            [j1map release];
            j1map = nil;
            joy1 = NULL;

            /* Reenable the old source on player 2's menu. */
            [[p2input itemAtIndex:[p1curinput tag] + 1] setEnabled:YES];

            /* Deselect the old input source. */
            [p1curinput setState:NSOffState];

            /* Select the keyboard in the menu. */
            [p1keyboard setState:NSOnState];
            p1curinput = p1keyboard;
        }
    }
    else    {
        /* Grab the specified joystick. */
        joy1 = joy_get_joystick(tag);

        /* Attempt to open the specified joystick. */
        if(!joy_open_joystick(joy1))    {
            /* The joystick has not been opened, show an error and keep the old
               selected input source. */
            NSString *title = [NSString stringWithFormat:@"Cannot open "
                "joystick %s.", joy1->name];
            NSRunAlertPanel(title, @"CrabEmu cannot open the specified "
                            "joystick. Verify that the joystick is attached "
                            "to the system and is powered on (if applicable).", 
                            @"OK", NULL, NULL);

            joy1 = old;
        }
        else    {
            /* Grab the button mapping for the specified joystick. */
            if(j1map != nil)    {
                [j1map release];
            }

            str = [NSString stringWithUTF8String:joy1->name];
            j1map = (NSMutableDictionary *)[[keymap objectForKey:str] retain];

            /* Close the old joystick, if any. */
            if(old) {
                joy_close_joystick(old);
            }

            /* Disable the joystick on the second player's menu. Reenable the
               old item. */
            [[p2input itemAtIndex:tag + 1] setEnabled:NO];
            [[p2input itemAtIndex:[p1curinput tag] + 1] setEnabled:YES];

            /* Set the old input source to the off state, and the new one to
               the on state. */
            [p1curinput setState:NSOffState];
            [item setState:NSOnState];
            p1curinput = item;
        }
    }

    /* We're done, unlock the memory lock. */
    [memoryLock unlock];
}

- (IBAction)setInput2:(id)sender
{
    NSMenuItem *item = (NSMenuItem *)sender;
    int tag;
    NSString *str;
    joydata_t *old = joy2;

    /* If we have tried to select the same input device we currently are using,
       ignore it. */
    if([item state] == NSOnState)   {
        return;
    }

    /* Lock the memory lock, and grab the tag of the menu item. */
    [memoryLock lock];
    tag = [item tag];

    /* If the user has selected the keyboard, close the old joytick if any,
       and release the old button mapping. */
    if(tag == -1)   {
        if(old) {
            /* Release resources associated with the joystick. */
            joy_close_joystick(old);
            [j2map release];
            j2map = nil;
            joy2 = NULL;

            /* Reenable the old source on player 1's menu. */
            [[p1input itemAtIndex:[p2curinput tag] + 1] setEnabled:YES];

            /* Deselect the old input source. */
            [p2curinput setState:NSOffState];

            /* Select the keyboard in the menu. */
            [p2keyboard setState:NSOnState];
            p2curinput = p2keyboard;
        }
    }
    else    {
        /* Grab the specified joystick. */
        joy2 = joy_get_joystick(tag);

        /* Attempt to open the specified joystick. */
        if(!joy_open_joystick(joy2))    {
            /* The joystick has not been opened, show an error and keep the old
               selected input source. */
            NSString *title = [NSString stringWithFormat:@"Cannot open "
                "joystick %s.", joy2->name];
            NSRunAlertPanel(title, @"CrabEmu cannot open the specified "
                            "joystick. Verify that the joystick is attached "
                            "to the system and is powered on (if applicable).", 
                            @"OK", NULL, NULL);

            joy2 = old;
        }
        else    {
            /* Grab the button mapping for the specified joystick. */
            if(j2map != nil)    {
                [j2map release];
            }

            str = [NSString stringWithUTF8String:joy2->name];
            j2map = (NSMutableDictionary *)[[keymap objectForKey:str] retain];

            /* Close the old joystick, if any. */
            if(old) {
                joy_close_joystick(old);
            }

            /* Disable the joystick on the first player's menu. Reenable the
               old item. */
            [[p1input itemAtIndex:tag + 1] setEnabled:NO];
            [[p1input itemAtIndex:[p2curinput tag] + 1] setEnabled:YES];

            /* Set the old input source to the off state, and the new one to
               the on state. */
            [p2curinput setState:NSOffState];
            [item setState:NSOnState];
            p2curinput = item;
        }
    }

    /* We're done, unlock the memory lock. */
    [memoryLock unlock];    
}

- (void)setJoystick:(joydata_t *)joy
             button:(int)button
         assignment:(uint32)a
{
    NSString *str = [NSString stringWithUTF8String:joy->name];
    NSMutableDictionary *d;

    if((d = (NSMutableDictionary *)[keymap objectForKey:str]) == nil)   {
        d = [NSMutableDictionary dictionaryWithCapacity:7];
        [keymap setObject:d forKey:str];
    }

    switch(button)  {
        case 0:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"SMS Up"];
            break;

        case 1:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"SMS Down"];
            break;

        case 2:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"SMS Left"];
            break;

        case 3:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"SMS Right"];
            break;

        case 4:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"SMS Button 1"];
            break;

        case 5:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"SMS Button 2"];
            break;

        case 6:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Game Gear Start"];
            break;
    }

    [self savePrefs];

    if([d count] >= 7)  {
        [self scanJoysticks:nil];
    }
}

- (void)joyInputThread:(id)b
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    NSButton *button = (NSButton *)b;
    int tag = [button tag];
    joydata_t *joy = joy_get_joystick([pjoysticks indexOfSelectedItem]);
    int i, anum, was_open;
    uint32 num = 0;

    /* Make sure the user can't hit anything until the sheet has opened. */
    usleep(50 * 1000);

    was_open = joy->open;

    if(!was_open)   {
        joy_open_joystick(joy);
    }

    while(!num) {
        for(i = 0; i < joy->buttons_count; ++i) {
            if(joy_read_button(joy, i)) {
                num = i;
            }
        }

        for(i = 0; i < joy->hats_count; ++i)    {
            anum = joy_read_hat(joy, i);

            if(anum != JOY_HAT_CENTER)  {
                num = 0x80000000 | (anum << 8) | i;
            }
        }

        for(i = 0; i < joy->axes_count; ++i)    {
            anum = joy_read_axis(joy, i);

            if(anum > 16384)    {
                num = 0x40000000 | i;
            }
            else if(anum < -16384)  {
                num = 0x60000000 | i;
            }
        }

        /* Sleep for 10ms */
        usleep(10 * 1000);
    }

    if(!was_open)   {
        joy_close_joystick(joy);
    }

    [self setJoystick:joy button:tag assignment:num];
    [NSApp abortModal];

	[pool release];
}

- (IBAction)prefJoyButton:(id)sender
{
    [NSThread detachNewThreadSelector:@selector(joyInputThread:)
                             toTarget:self
                           withObject:sender];

    [NSApp beginSheet:joyPanel
       modalForWindow:prefWindow
        modalDelegate:self
       didEndSelector:NULL
          contextInfo:NULL];

    [NSApp runModalForWindow:joyPanel];

    [NSApp endSheet:joyPanel];
    [joyPanel orderOut:nil];
}

- (IBAction)prefSMSBiosBrowse:(id)sender
{
	NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    NSArray *fileTypes = [NSArray arrayWithObjects:@"sms", nil];
	[openPanel setTitle:@"Choose a Master System BIOS Rom"];
	[openPanel setCanChooseDirectories:NO];

	if([openPanel runModalForTypes:fileTypes] == NSOKButton)	{
		NSString *fn = [openPanel filename];

        [smsBios setStringValue:fn];
        [prefs setObject:fn forKey:@"SMS BIOS ROM"];
        [prefs synchronize];
    }
}

- (IBAction)prefGGBiosBrowse:(id)sender
{
	NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    NSArray *fileTypes = [NSArray arrayWithObjects:@"gg", nil];
	[openPanel setTitle:@"Choose a Game Gear BIOS Rom"];
	[openPanel setCanChooseDirectories:NO];

	if([openPanel runModalForTypes:fileTypes] == NSOKButton)	{
		NSString *fn = [openPanel filename];

        [ggBios setStringValue:fn];
        [prefs setObject:fn forKey:@"Game Gear BIOS ROM"];
        [prefs synchronize];
    }
}

- (IBAction)prefBiosEnabledClicked:(id)sender
{
    [self savePrefs];
}

- (IBAction)takeScreenshot:(id)sender
{
    NSSavePanel *panel;
    NSBitmapImageRep *img;
    int w, h, i, j, offh, offw, idx;
    unsigned char *data;

    if(!sms_initialized)
        return;

    [memoryLock lock];

    panel = [NSSavePanel savePanel];
    [panel setTitle:@"Save Screenshot As"];
    [panel setRequiredFileType:@"png"];

    if([panel runModal] != NSOKButton)  {
        [memoryLock unlock];
        return;
    }

    if(sms_console == CONSOLE_GG)   {
        w = 160;
        h = 144;
        offh = 24;
        offw = 48;
    }
    else    {
        w = 256;
        h = smsvdp.lines;
        offh = offw = 0;
    }

    w = sms_console == CONSOLE_GG ? 160 : 256;
    h = sms_console == CONSOLE_GG ? 144 : smsvdp.lines;

    img = [[NSBitmapImageRep alloc]
            initWithBitmapDataPlanes:NULL
                          pixelsWide:w
                          pixelsHigh:h
                       bitsPerSample:8
                     samplesPerPixel:3
                            hasAlpha:NO
                            isPlanar:NO
                      colorSpaceName:NSDeviceRGBColorSpace
                         bytesPerRow:0
                        bitsPerPixel:32];

    if(img == nil)  {
        NSRunAlertPanel(@"Error", @"CrabEmu cannot create a NSBitmapImageRep"
                        "for taking a screenshot. Please file a bug report.",
                        @"OK", NULL, NULL);
        [memoryLock unlock];
        return;
    }

    data = [img bitmapData];

    for(i = 0; i < h; ++i)  {
        for(j = 0; j < w; ++j)  {
            idx = (i + offh) * 256 + j + offw;

            data[(i * w + j) * 4] = (smsvdp.framebuffer[idx] >> 16) & 0xFF;
            data[(i * w + j) * 4 + 1] = (smsvdp.framebuffer[idx] >> 8) & 0xFF;
            data[(i * w + j) * 4 + 2] = smsvdp.framebuffer[idx] & 0xFF;
            data[(i * w + j) * 4 + 3] = 0xFF;
        }
    }

    if(![[img representationUsingType:NSPNGFileType
                           properties:NULL] writeToFile:[panel filename]
                                             atomically:NO])    {
        NSRunAlertPanel(@"Error", @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to"
                        "this directory, and this file.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)toggleFullscreen:(id)sender
{
    if(!sms_initialized)
        return;

    [theview toggleFullscreen];
}

- (IBAction)toggleScreensaver:(id)sender
{
    if([ssaverDisabled state] == NSOnState) {
        if(timer == nil)    {
            /* The user has newly selected to disable the screensaver, start
               the timer. */
            timer = [NSTimer scheduledTimerWithTimeInterval:15
                                                     target:self
                                                   selector:@selector(ssTimer:)
                                                   userInfo:nil
                                                    repeats:YES];
        }
    }
    else    {
        if(timer != nil)    {
            /* The user has selected to not disable the screensaver, invalidate
               the existing timer, and set it to nil. The run loop will
               release the object for us. */
            [timer invalidate];
            timer = nil;
        }
    }

    [self savePrefs];
}

- (IBAction)showVDPPalette:(id)sender
{
    if(!sms_initialized)
        return;

    if([showVDPPalette state] == NSOnState) {
        [showVDPPalette setState:NSOffState];
        [vdpPalettePanel orderOut:self];
    }
    else    {
        [showVDPPalette setState:NSOnState];
        [vdpPalettePanel orderFront:self];
    }
}

- (IBAction)showVDPPatterns:(id)sender
{
    if(!sms_initialized)
        return;

    if([showVDPPatterns state] == NSOnState)    {
        [showVDPPatterns setState:NSOffState];
        [vdpPatternPanel orderOut:self];
    }
    else    {
        [showVDPPatterns setState:NSOnState];
        [vdpPatternPanel orderFront:self];
    }
} 

- (BOOL)initMasterSystem
{
    int i;

    if(sms_initialized == 1)    {
        sms_write_cartram_to_file();
        if(sms_reset())
            return NO;
    }
    else    {
        if(sms_init(ntscVideo == YES ? SMS_VIDEO_NTSC : SMS_VIDEO_PAL,
                    domestic == YES ? SMS_REGION_DOMESTIC : SMS_REGION_EXPORT))
            return NO;
    }

    if([smsBiosEnabled state] == NSOnState &&
       [[smsBios stringValue] length] > 0)  {
        sms_mem_load_bios([[smsBios stringValue] UTF8String]);
    }

    if([ggBiosEnabled state] == NSOnState &&
       [[ggBios stringValue] length] > 0)   {
        sms_mem_load_bios([[ggBios stringValue] UTF8String]);
    }

    [theview showWindow];

    if(started == 0)    {
        [NSThread detachNewThreadSelector:@selector(masterEmuThread:)
                                 toTarget:self
                               withObject:nil];
        started = 1;
    }

    /* Enable everything that needs to be enabled while a ROM is running... */
    [screenshot setEnabled:YES];

    if(CGLLockContext != NULL)  {
        [fullscreen setEnabled:YES];
    }

    [saveSRAM setEnabled:YES];

    for(i = 0; i < [saveStates numberOfItems]; ++i) {
        [[saveStates itemAtIndex:i] setEnabled:YES];
    }
    
    for(i = 0; i < [loadStates numberOfItems]; ++i) {
        [[loadStates itemAtIndex:i] setEnabled:YES];
    }

    [pause setEnabled:YES];
    [reset setEnabled:YES];

    /* ...and disable those settings we don't want available while a ROM is
       running. */
    [ntsc setEnabled:NO];
    [pal setEnabled:NO];
    [rExport setEnabled:NO];
    [rDomestic setEnabled:NO];

    return YES;
}

- (BOOL)loadMasterSystemRomFromFile:(NSString *)filename
{
    if(openRom != nil)  {
        [openRom release];
    }

    openRom = [filename retain];

    if(sms_mem_load_rom([filename UTF8String])) {
        [openRom release];
        openRom = nil;

        return NO;
    }

    if(sms_console == CONSOLE_SMS && [[smsBios stringValue] length] > 0 &&
       [smsBiosEnabled state] == NSOnState) {
        sms_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                              SMS_MEMCTL_RAM) ^ 0xFF);
    }
    else if(sms_console == CONSOLE_GG && [[ggBios stringValue] length] > 0 &&
            [ggBiosEnabled state] == NSOnState) {
        sms_gg_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                                 SMS_MEMCTL_CART | SMS_MEMCTL_RAM) ^ 0xFF);
    }

    return YES;
}

- (void)fillDictionary:(NSMutableDictionary *)d withButtons:(unichar *)string
{
    [d setObject:[NSNumber numberWithInt:string[0]]
          forKey:@"SMS Up"];
    [d setObject:[NSNumber numberWithInt:string[1]]
          forKey:@"SMS Down"];
    [d setObject:[NSNumber numberWithInt:string[2]]
          forKey:@"SMS Left"];
    [d setObject:[NSNumber numberWithInt:string[3]]
          forKey:@"SMS Right"];
    [d setObject:[NSNumber numberWithInt:string[4]]
          forKey:@"SMS Button 1"];
    [d setObject:[NSNumber numberWithInt:string[5]]
          forKey:@"SMS Button 2"];
    [d setObject:[NSNumber numberWithInt:string[6]]
          forKey:@"Game Gear Start"];
}

- (void)fillString:(unichar *)string withDictionary:(NSDictionary *)d
{
    NSNumber *n;

    n = (NSNumber *)[d objectForKey:@"SMS Up"];
    string[0] = (unichar)[n intValue];
    n = (NSNumber *)[d objectForKey:@"SMS Down"];
    string[1] = (unichar)[n intValue];
    n = (NSNumber *)[d objectForKey:@"SMS Left"];
    string[2] = (unichar)[n intValue];
    n = (NSNumber *)[d objectForKey:@"SMS Right"];
    string[3] = (unichar)[n intValue];
    n = (NSNumber *)[d objectForKey:@"SMS Button 1"];
    string[4] = (unichar)[n intValue];
    n = (NSNumber *)[d objectForKey:@"SMS Button 2"];
    string[5] = (unichar)[n intValue];
    n = (NSNumber *)[d objectForKey:@"Game Gear Start"];
    string[6] = (unichar)[n intValue];
}

- (void)savePrefs
{
    unichar string[14];
    NSMutableDictionary *keys;

    [buttonmap getCharacters:string];

    keys = [[NSMutableDictionary dictionaryWithCapacity:7] retain];
    [self fillDictionary:keys withButtons:string];
    [keymap setObject:keys forKey:@"Keyboard 1"];
    [keys release];

    keys = [[NSMutableDictionary dictionaryWithCapacity:7] retain];
    [self fillDictionary:keys withButtons:string + 7];
    [keymap setObject:keys forKey:@"Keyboard 2"];
    [keys release];

    /* Save both 0.1.9 and higher style button mappings and 0.1.8 and lower
       style mappings, just in case someone actually has two different versions
       laying around. */
    [prefs setObject:buttonmap forKey:@"SMS Buttons"];
    [prefs setObject:keymap forKey:@"SMS Buttonmap"];

    [prefs setBool:!ntscVideo forKey:@"SMS PAL"];
    [prefs setBool:domestic forKey:@"SMS Domestic"];

    [prefs setBool:[smsBiosEnabled state] == NSOnState
            forKey:@"SMS BIOS Enabled"];
    [prefs setBool:[ggBiosEnabled state] == NSOnState
            forKey:@"GG BIOS Enabled"];

    [prefs setBool:[ssaverDisabled state] == NSOnState
            forKey:@"Disable Screensaver"];

    [prefs synchronize];
}

- (void)ssTimer:(NSTimer *)timer
{
    UpdateSystemActivity(UsrActivity);
}

@end
