/*
    This file is part of CrabEmu.

    Copyright (C) 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <QApplication>
#include <QMessageBox>
#include <QFileDialog>
#include <QKeyEvent>
#include <QSettings>

#include "UICrabEmu.h"
#include "UIAboutDialog.h"
#include "UIPreferences.h"
#include "../EmuWidget.h"
#include "sms.h"
#include "smsmem.h"
#include "smsmem-gg.h"
#include "smsvdp.h"
#include "sn76489.h"
#include "smsz80.h"
#include "sound.h"

extern "C" int sms_initialized;
extern "C" int sms_psg_enabled;
extern "C" sn76489_t psg;
extern "C" int sms_console;
extern QMutex memoryLock;

UICrabEmu::UICrabEmu(QWidget *parent) : QMainWindow(parent) {
    QSettings settings;
    bool pal, domestic;

    setupUi(this);
    setAttribute(Qt::WA_DeleteOnClose);

    thd = new CrabEmuThread;
    thd->start();

    emuwidget = new EmuWidget(thd, this);
    setCentralWidget(emuwidget);
    emuwidget->setFocusProxy(this);

    paused = false;

    /* If the user doesn't currently have any settings, set up the defaults. */
    if(!settings.allKeys().contains("SMS Buttonmap"))   {
        initializeSettings();
        settings.sync();
    }

    pal = settings.value("SMS PAL").toBool();
    domestic = settings.value("SMS Domestic").toBool();

    videoSystem = pal ? SMS_VIDEO_PAL : SMS_VIDEO_NTSC;
    actionNTSC->setChecked(!pal);
    actionPAL->setChecked(pal);

    region = domestic ? SMS_REGION_DOMESTIC : SMS_REGION_EXPORT;
    actionExport->setChecked(!domestic);
    actionDomestic->setChecked(domestic);
}

CrabEmuThread *UICrabEmu::thread()  {
    return thd;
}

void UICrabEmu::on_actionOpenRom_triggered()    {
    QString fn;
    QSettings settings;
    QString biosfn;

    fn = QFileDialog::getOpenFileName(QApplication::activeWindow(),
                                      tr("Open Rom"), "",
                                      tr("Roms (*.sc *.sg *.sms *.gg)"), 0,
                                      QFileDialog::DontUseSheet);

    /* If a file was not selected, then don't try to open one. */
    if(fn.isEmpty())    {
        return;
    }

    openRom = fn;

    memoryLock.lock();

    if(sms_initialized) {
        sms_write_cartram_to_file();
        sms_reset();
    }
    else    {
        sms_init(videoSystem, region);
    }

    sms_mem_load_rom(fn.toUtf8().data());

    if(sms_console == CONSOLE_SMS)  {
        biosfn = settings.value("SMS BIOS ROM").toString();
        if(settings.value("SMS BIOS Enabled", false).toBool() &&
           biosfn.length()) {
            sms_mem_load_bios(biosfn.toUtf8().data());
            sms_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                                  SMS_MEMCTL_RAM) ^ 0xFF);
        }
    }
    else if(sms_console == CONSOLE_GG)  {
        biosfn = settings.value("Game Gear BIOS ROM").toString();
        if(settings.value("GG BIOS Enabled", false).toBool() &&
           biosfn.length()) {
            sms_mem_load_bios(biosfn.toUtf8().data());
            sms_gg_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                                     SMS_MEMCTL_CART | SMS_MEMCTL_RAM) ^ 0xFF);
        }
    }

    memoryLock.unlock();
}

void UICrabEmu::on_actionSaveSRAM_triggered()   {
    if(sms_initialized)
        sms_write_cartram_to_file();
}

void UICrabEmu::on_actionSaveSlot1_triggered()  {
    QString filename;

    if(sms_initialized == 0)
        return;

    filename = openRom + ".1.cestate";

    memoryLock.lock();

    if(sms_save_state(filename.toUtf8().data()))    {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open save "
                             "state file", "CrabEmu cannot open this file for "
                             "writing. Verify that you have the permissions to "
                             "write to this directory and that the target "
                             "volume is not full.");
    }

    memoryLock.unlock();
}

void UICrabEmu::on_actionSaveSlot2_triggered()  {
    QString filename;

    if(sms_initialized == 0)
        return;

    filename = openRom + ".2.cestate";

    memoryLock.lock();

    if(sms_save_state(filename.toUtf8().data()))    {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open save "
                             "state file", "CrabEmu cannot open this file for "
                             "writing. Verify that you have the permissions to "
                             "write to this directory and that the target "
                             "volume is not full.");
    }

    memoryLock.unlock();
}

void UICrabEmu::on_actionSaveSlot3_triggered()  {
    QString filename;

    if(sms_initialized == 0)
        return;

    filename = openRom + ".3.cestate";

    memoryLock.lock();

    if(sms_save_state(filename.toUtf8().data()))    {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open save "
                             "state file", "CrabEmu cannot open this file for "
                             "writing. Verify that you have the permissions to "
                             "write to this directory and that the target "
                             "volume is not full.");
    }

    memoryLock.unlock();
}

void UICrabEmu::on_actionSaveSlot4_triggered()  {
    QString filename;

    if(sms_initialized == 0)
        return;

    filename = openRom + ".4.cestate";

    memoryLock.lock();

    if(sms_save_state(filename.toUtf8().data()))    {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open save "
                             "state file", "CrabEmu cannot open this file for "
                             "writing. Verify that you have the permissions to "
                             "write to this directory and that the target "
                             "volume is not full.");
    }

    memoryLock.unlock();
}

void UICrabEmu::on_actionSaveSlot5_triggered()  {
    QString filename;

    if(sms_initialized == 0)
        return;

    filename = openRom + ".5.cestate";

    memoryLock.lock();

    if(sms_save_state(filename.toUtf8().data()))    {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open save "
                             "state file", "CrabEmu cannot open this file for "
                             "writing. Verify that you have the permissions to "
                             "write to this directory and that the target "
                             "volume is not full.");
    }

    memoryLock.unlock();
}

void UICrabEmu::on_actionSaveSlot6_triggered()  {
    QString filename;

    if(sms_initialized == 0)
        return;

    filename = openRom + ".6.cestate";

    memoryLock.lock();

    if(sms_save_state(filename.toUtf8().data()))    {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open save "
                             "state file", "CrabEmu cannot open this file for "
                             "writing. Verify that you have the permissions to "
                             "write to this directory and that the target "
                             "volume is not full.");
    }

    memoryLock.unlock();
}

void UICrabEmu::on_actionLoadSlot1_triggered()  {
    QString filename;

    if(sms_initialized == 0)
        return;

    filename = openRom + ".1.cestate";

    memoryLock.lock();

    if(sms_load_state(filename.toUtf8().data()))    {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open save "
                             "state file", "CrabEmu cannot open this file for "
                             "reading. Verify that you have the permissions to "
                             "read this directory and that the file exists.");
    }

    memoryLock.unlock();
}

void UICrabEmu::on_actionLoadSlot2_triggered()  {
    QString filename;

    if(sms_initialized == 0)
        return;

    filename = openRom + ".2.cestate";

    memoryLock.lock();

    if(sms_load_state(filename.toUtf8().data()))    {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open save "
                             "state file", "CrabEmu cannot open this file for "
                             "reading. Verify that you have the permissions to "
                             "read this directory and that the file exists.");
    }

    memoryLock.unlock();
}

void UICrabEmu::on_actionLoadSlot3_triggered()  {
    QString filename;

    if(sms_initialized == 0)
        return;

    filename = openRom + ".3.cestate";

    memoryLock.lock();

    if(sms_load_state(filename.toUtf8().data()))    {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open save "
                             "state file", "CrabEmu cannot open this file for "
                             "reading. Verify that you have the permissions to "
                             "read this directory and that the file exists.");
    }

    memoryLock.unlock();
}

void UICrabEmu::on_actionLoadSlot4_triggered()  {
    QString filename;

    if(sms_initialized == 0)
        return;

    filename = openRom + ".4.cestate";

    memoryLock.lock();

    if(sms_load_state(filename.toUtf8().data()))    {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open save "
                             "state file", "CrabEmu cannot open this file for "
                             "reading. Verify that you have the permissions to "
                             "read this directory and that the file exists.");
    }

    memoryLock.unlock();
}

void UICrabEmu::on_actionLoadSlot5_triggered()  {
    QString filename;

    if(sms_initialized == 0)
        return;

    filename = openRom + ".5.cestate";

    memoryLock.lock();

    if(sms_load_state(filename.toUtf8().data()))    {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open save "
                             "state file", "CrabEmu cannot open this file for "
                             "reading. Verify that you have the permissions to "
                             "read this directory and that the file exists.");
    }

    memoryLock.unlock();
}

void UICrabEmu::on_actionLoadSlot6_triggered()  {
    QString filename;

    if(sms_initialized == 0)
        return;

    filename = openRom + ".6.cestate";

    memoryLock.lock();

    if(sms_load_state(filename.toUtf8().data()))    {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open save "
                             "state file", "CrabEmu cannot open this file for "
                             "reading. Verify that you have the permissions to "
                             "read this directory and that the file exists.");
    }

    memoryLock.unlock();
}

void UICrabEmu::on_actionTakeScreenshot_triggered() {
    QMutexLocker locker(&memoryLock);

    if(!sms_initialized)    {
        return;
    }

    QImage *data;
    QImage img;
    QString fn = QFileDialog::getSaveFileName(QApplication::activeWindow(),
                                              tr("Save Screenshot As"), "",
                                              tr("PNG Images (*.png)"), 0,
                                              QFileDialog::DontUseSheet);

    if(fn.isEmpty())    {
        return;
    }

    if(!fn.endsWith(".png", Qt::CaseInsensitive))   {
        fn += ".png";
    }

    data = emuwidget->retainImage();

    if(sms_console == CONSOLE_GG)   {
        img = data->copy(48, 24, 160, 144);
    }
    else    {
        img = data->copy(0, 0, 256, smsvdp.lines);
    }

    img.save(fn, "PNG");

    emuwidget->releaseImage();
}

void UICrabEmu::on_actionExit_triggered()   {
    thd->finish();
    thd->wait();
    delete thd;

    if(sms_initialized) {
        sms_write_cartram_to_file();
        sms_shutdown();
    }

    close();
}

void UICrabEmu::on_actionEnablePSG_triggered()  {
    if(!actionEnablePSG->isChecked())   {
        sms_psg_enabled = 0;
    }
    else    {
        sms_psg_enabled = 1;
    }
}

void UICrabEmu::on_actionTone0_triggered()  {
    if(!actionTone0->isChecked())   {
        psg.enabled_channels &= ~ENABLE_TONE0;
    }
    else    {
        psg.enabled_channels |= ENABLE_TONE0;
    }
}

void UICrabEmu::on_actionTone1_triggered()  {
    if(!actionTone1->isChecked())   {
        psg.enabled_channels &= ~ENABLE_TONE1;
    }
    else    {
        psg.enabled_channels |= ENABLE_TONE1;
    }
}

void UICrabEmu::on_actionTone2_triggered()  {
    if(!actionTone2->isChecked())   {
        psg.enabled_channels &= ~ENABLE_TONE2;
    }
    else    {
        psg.enabled_channels |= ENABLE_TONE2;
    }
}

void UICrabEmu::on_actionNoise_triggered()  {
    if(!actionNoise->isChecked())   {
        psg.enabled_channels &= ~ENABLE_NOISE;
    }
    else    {
        psg.enabled_channels |= ENABLE_NOISE;
    }
}

void UICrabEmu::on_actionExport_triggered() {
    if(!actionExport->isChecked())  {
        actionExport->setChecked(true);
        return;
    }

    actionDomestic->setChecked(false);
    region = SMS_REGION_EXPORT;

    QSettings settings;
    settings.setValue("SMS Domestic", false);
}

void UICrabEmu::on_actionDomestic_triggered()   {
    if(!actionDomestic->isChecked())    {
        actionDomestic->setChecked(true);
        return;
    }

    actionExport->setChecked(false);
    region = SMS_REGION_DOMESTIC;

    QSettings settings;
    settings.setValue("SMS Domestic", true);
}

void UICrabEmu::on_actionNTSC_triggered()   {
    if(!actionNTSC->isChecked())  {
        actionNTSC->setChecked(true);
        return;
    }

    actionPAL->setChecked(false);
    videoSystem = SMS_VIDEO_NTSC;

    QSettings settings;
    settings.setValue("SMS PAL", false);
}

void UICrabEmu::on_actionPAL_triggered()    {
    if(!actionPAL->isChecked()) {
        actionPAL->setChecked(true);
        return;
    }

    actionNTSC->setChecked(false);
    videoSystem = SMS_VIDEO_PAL;

    QSettings settings;
    settings.setValue("SMS PAL", true);
}

void UICrabEmu::on_actionPause_triggered()  {
    if(!sms_initialized)
        return;

    if(sms_console == CONSOLE_SMS)  {
        memoryLock.lock();
        sms_z80_nmi();
        memoryLock.unlock();
    }
    else if(!paused)    {
        memoryLock.lock();
        sound_pause();
        paused = true;
    }
    else    {
        paused = false;
        sound_unpause();
        memoryLock.unlock();
    }
}

void UICrabEmu::on_actionReset_triggered()  {
    if(!sms_initialized)
        return;

    sms_soft_reset();
}

void UICrabEmu::on_actionPreferences_triggered()    {
    UIPreferences prefs(window());
    int result = prefs.exec();

    if(result == QDialog::Accepted) {
        prefs.updatePreferences();
    }
}

void UICrabEmu::on_actionAbout_triggered()  {
    UIAboutDialog(window()).exec();
}

void UICrabEmu::keyPressEvent(QKeyEvent *e) {
    int button = 0;
    QSettings settings;
    QVariantMap buttonmap, kb1, kb2;

    /* Grab the button maps from the settings files. */
    buttonmap = settings.value("SMS Buttonmap").toMap();
    kb1 = buttonmap.value("Qt Keyboard 1").toMap();
    kb2 = buttonmap.value("Qt Keyboard 2").toMap();

    /* Check the key in the event against each one in the 1st player's keyboard
       mapping. */
    if(e->key() == kb1.value("SMS Up").toInt()) {
        button = SMS_PAD1_UP;
    }
    else if(e->key() == kb1.value("SMS Down").toInt())  {
        button = SMS_PAD1_DOWN;
    }
    else if(e->key() == kb1.value("SMS Left").toInt())  {
        button = SMS_PAD1_LEFT;
    }
    else if(e->key() == kb1.value("SMS Right").toInt()) {
        button = SMS_PAD1_RIGHT;
    }
    else if(e->key() == kb1.value("SMS Button 1").toInt())  {
        button = SMS_PAD1_A;
    }
    else if(e->key() == kb1.value("SMS Button 2").toInt())  {
        button = SMS_PAD1_B;
    }
    else if(e->key() == kb1.value("Game Gear Start").toInt())   {
        button = GG_START;
    }

    if(button != 0) {
        sms_button_pressed(button);
        return;
    }

    /* Check the key in the event against each one in the 2nd player's keyboard
       mapping. */
    if(e->key() == kb2.value("SMS Up").toInt()) {
        button = SMS_PAD2_UP;
    }
    else if(e->key() == kb2.value("SMS Down").toInt())  {
        button = SMS_PAD2_DOWN;
    }
    else if(e->key() == kb2.value("SMS Left").toInt())  {
        button = SMS_PAD2_LEFT;
    }
    else if(e->key() == kb2.value("SMS Right").toInt()) {
        button = SMS_PAD2_RIGHT;
    }
    else if(e->key() == kb2.value("SMS Button 1").toInt())  {
        button = SMS_PAD2_A;
    }
    else if(e->key() == kb2.value("SMS Button 2").toInt())  {
        button = SMS_PAD2_B;
    }

    if(button != 0) {
        sms_button_pressed(button);
    }
}

void UICrabEmu::keyReleaseEvent(QKeyEvent *e)   {
    int button = 0;
    QSettings settings;
    QVariantMap buttonmap, kb1, kb2;

    /* Grab the button maps from the settings files. */
    buttonmap = settings.value("SMS Buttonmap").toMap();
    kb1 = buttonmap.value("Qt Keyboard 1").toMap();
    kb2 = buttonmap.value("Qt Keyboard 2").toMap();

    /* Check the key in the event against each one in the 1st player's keyboard
       mapping. */
    if(e->key() == kb1.value("SMS Up").toInt()) {
        button = SMS_PAD1_UP;
    }
    else if(e->key() == kb1.value("SMS Down").toInt())  {
        button = SMS_PAD1_DOWN;
    }
    else if(e->key() == kb1.value("SMS Left").toInt())  {
        button = SMS_PAD1_LEFT;
    }
    else if(e->key() == kb1.value("SMS Right").toInt()) {
        button = SMS_PAD1_RIGHT;
    }
    else if(e->key() == kb1.value("SMS Button 1").toInt())  {
        button = SMS_PAD1_A;
    }
    else if(e->key() == kb1.value("SMS Button 2").toInt())  {
        button = SMS_PAD1_B;
    }
    else if(e->key() == kb1.value("Game Gear Start").toInt())   {
        button = GG_START;
    }

    if(button != 0) {
        sms_button_released(button);
        return;
    }

    /* Check the key in the event against each one in the 2nd player's keyboard
       mapping. */
    if(e->key() == kb2.value("SMS Up").toInt()) {
        button = SMS_PAD2_UP;
    }
    else if(e->key() == kb2.value("SMS Down").toInt())  {
        button = SMS_PAD2_DOWN;
    }
    else if(e->key() == kb2.value("SMS Left").toInt())  {
        button = SMS_PAD2_LEFT;
    }
    else if(e->key() == kb2.value("SMS Right").toInt()) {
        button = SMS_PAD2_RIGHT;
    }
    else if(e->key() == kb2.value("SMS Button 1").toInt())  {
        button = SMS_PAD2_A;
    }
    else if(e->key() == kb2.value("SMS Button 2").toInt())  {
        button = SMS_PAD2_B;
    }

    if(button != 0) {
        sms_button_released(button);
    }
}

void UICrabEmu::initializeSettings()    {
    QSettings settings;
    QVariantMap buttonmap, kb1, kb2;

    settings.setValue("SMS BIOS ROM", "");
    settings.setValue("SMS Domestic", false);
    settings.setValue("SMS PAL", false);

    kb1["SMS Up"] = Qt::Key_Up;
    kb1["SMS Down"] = Qt::Key_Down;
    kb1["SMS Left"] = Qt::Key_Left;
    kb1["SMS Right"] = Qt::Key_Right;
    kb1["SMS Button 1"] = Qt::Key_Semicolon;
    kb1["SMS Button 2"] = Qt::Key_Apostrophe;
    kb1["Game Gear Start"] = Qt::Key_P;

    kb2["SMS Up"] = Qt::Key_W;
    kb2["SMS Down"] = Qt::Key_S;
    kb2["SMS Left"] = Qt::Key_A;
    kb2["SMS Right"] = Qt::Key_D;
    kb2["SMS Button 1"] = Qt::Key_V;
    kb2["SMS Button 2"] = Qt::Key_B;
    kb2["Game Gear Start"] = Qt::Key_R;

    buttonmap["Qt Keyboard 1"] = kb1;
    buttonmap["Qt Keyboard 2"] = kb2;
    settings.setValue("SMS Buttonmap", buttonmap);
}
